VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CTaskList"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' *************************************************************************
'  Copyright 1998-2000 Karl E. Peterson
'  All Rights Reserved, http://www.mvps.org/vb
' *************************************************************************
'  You are free to use this code within your own applications, but you
'  are expressly forbidden from selling or otherwise distributing this
'  source code, non-compiled, without prior written consent.
' *************************************************************************
Option Explicit

Private Declare Function GetTopWindow Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function GetNextWindow Lib "user32" Alias "GetWindow" (ByVal hWnd As Long, ByVal wFlag As Long) As Long
Private Declare Function GetWindowText Lib "user32" Alias "GetWindowTextA" (ByVal hWnd As Long, ByVal lpString As String, ByVal cch As Long) As Long
Private Declare Function GetWindowLong Lib "user32" Alias "GetWindowLongA" (ByVal hWnd As Long, ByVal nIndex As Long) As Long
Private Declare Function IsWindow Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function IsWindowVisible Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function GetClassname Lib "user32" Alias "GetClassNameA" (ByVal hWnd As Long, ByVal lpClassName As String, ByVal nMaxCount As Long) As Long
Private Declare Function GetClassLong Lib "user32" Alias "GetClassLongA" (ByVal hWnd As Long, ByVal nIndex As Long) As Long
Private Declare Function GetParent Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long
Private Declare Function SendMessageTimeout Lib "user32" Alias "SendMessageTimeoutA" (ByVal hWnd As Long, ByVal msg As Long, ByVal wParam As Long, lParam As Any, ByVal fuFlags As Long, ByVal uTimeout As Long, lpdwResult As Long) As Long
Private Declare Function IsIconic Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function ShowWindow Lib "user32" (ByVal hWnd As Long, ByVal nCmdShow As Long) As Long
Private Declare Function LoadIcon Lib "user32" Alias "LoadIconA" (ByVal hInstance As Long, lpIconName As Any) As Long
Private Declare Function GetForegroundWindow Lib "user32" () As Long
Private Declare Function SetForegroundWindow Lib "user32" (ByVal hWnd As Long) As Long

Private Const GW_HWNDFIRST = 0
Private Const GW_HWNDNEXT = 2

Private Const ICON_SMALL = 0
Private Const ICON_BIG = 1
Private Const GCL_HICON = (-14)
Private Const GCL_HICONSM = (-34)

Private Const GWL_STYLE = (-16)
Private Const GWL_EXSTYLE = (-20)
Private Const GWL_HWNDPARENT = (-8)
Private Const WS_EX_TOOLWINDOW = &H80

Private Const WM_NULL = &H0
Private Const WM_GETTEXT = &HD
Private Const WM_GETTEXTLENGTH = &HE
Private Const WM_GETICON = &H7F
Private Const WM_SETICON = &H80

Private Const SMTO_NORMAL = &H0
Private Const SMTO_BLOCK = &H1
Private Const SMTO_ABORTIFHUNG = &H2

Private Const SW_RESTORE = 9

Private Const IDI_WINLOGO = 32517&
Private Const IDI_APPLICATION = 32512&

Private Const ERROR_TIMEOUT = 1460

' Private storage for class data
Private Type TaskListItems
   hWnd As Long
   hIcon As Long
   hIconSm As Long
   Caption As String
End Type

Private m_Count As Long
Private m_tasks() As TaskListItems
Private m_hFlagIcon As Long
Private m_hAppIcon As Long
Private m_Mode As TaskListModes
Private m_HungTimeout As Long

Private Const defHungTimeout = 500 'milliseconds
Private Const ArrayBump = 20

' Operating modes:
Public Enum TaskListModes
   tlAltTab = 0
   tlTaskMgr = 1
End Enum

' Status values:
Public Enum TaskStatus
   tlStatusOK = 0
   tlStatusHung = 1
End Enum
   
' ************************************************
'  Initialize/Terminate
' ************************************************
Private Sub Class_Initialize()
   ' Grab handles to default icons.
   m_hFlagIcon = LoadIcon(0, ByVal IDI_WINLOGO)
   m_hAppIcon = LoadIcon(0, ByVal IDI_APPLICATION)
   ' Assign default class properties
   m_Mode = tlAltTab
   ' Set default timeout for "hung" tasks
   m_HungTimeout = defHungTimeout
   ' Initialize cache array
   ReDim m_tasks(0 To ArrayBump - 1) As TaskListItems
End Sub

Private Sub Class_Terminate()
   '
End Sub

' ************************************************
'  Public Properties (Read/Write)
' ************************************************
Public Property Let HungTimeout(ByVal NewVal As Long)
   ' Set new timeout period to consider
   ' a task "hung".
   If NewVal > 0 Then
      m_HungTimeout = NewVal
   End If
End Property

Public Property Get HungTimeout() As Long
   HungTimeout = m_HungTimeout
End Property

Public Property Let Mode(ByVal NewVal As TaskListModes)
   Select Case NewVal
      Case tlAltTab, tlTaskMgr
         m_Mode = NewVal
      Case Else
         ' Potentially raise an error, or
         ' just ignore bad input.
   End Select
End Property

Public Property Get Mode() As TaskListModes
   Mode = m_Mode
End Property

' ************************************************
'  Public Properties (Read-Only)
' ************************************************
Public Property Get Caption(ByVal Index As Long) As String
   If Index > 0 And Index <= m_Count Then
      ' Return associated caption.
      Caption = m_tasks(Index - 1).Caption
   End If
End Property

Public Property Get Count() As Long
   ' Return task count.
   Count = m_Count
End Property

Public Property Get hWnd(ByVal Index As Long) As Long
   If Index > 0 And Index <= m_Count Then
      ' Return handle of top-level window.
      hWnd = m_tasks(Index - 1).hWnd
   End If
End Property

Public Property Get hIcon(ByVal Index As Long) As Long
   If Index > 0 And Index <= m_Count Then
      If m_tasks(Index - 1).hIcon <> 0 Then
         ' Return extracted large icon.
         hIcon = m_tasks(Index - 1).hIcon
      Else
         ' No available icon, return suitable
         ' generic icon based on current mode.
         Select Case m_Mode
            Case tlAltTab
               hIcon = m_hFlagIcon
            Case tlTaskMgr
               hIcon = m_hAppIcon
         End Select
      End If
   End If
End Property

Public Property Get hIconSm(ByVal Index As Long) As Long
   If Index > 0 And Index <= m_Count Then
      ' Return extracted small icon if available,
      ' or large icon if small is not available.
      If m_tasks(Index - 1).hIconSm <> 0 Then
         hIconSm = m_tasks(Index - 1).hIconSm
      Else
         hIconSm = Me.hIcon(Index)
      End If
   End If
End Property

Public Property Get Status(ByVal Index As Long) As TaskStatus
   Static nRet As Long
   Static lpdwResult As Long
   Const flags As Long = SMTO_ABORTIFHUNG 'Or SMTO_BLOCK
   ' Detection of "hung" status based on KB article Q231844:
   ' http://support.microsoft.com/support/kb/articles/Q231/8/44.ASP
   If Index > 0 And Index <= m_Count Then
      ' Assume task is not hung.
      Status = tlStatusOK
      ' See if task responds to null msg.
      nRet = SendMessageTimeout(m_tasks(Index - 1).hWnd, WM_NULL, 0&, 0&, flags, m_HungTimeout, lpdwResult)
      If nRet = 0 Then
         ' In NT, Err.LastDllError will be
         ' set to ERROR_TIMEOUT, but we can't
         ' count on that in Win9x.
         Status = tlStatusHung
      End If
   End If
End Property

Public Property Get Valid(Optional ByVal Index As Long, Optional ByVal hWnd As Long) As Boolean
   Dim i As Long
   
   ' If passed an hWnd, find index within tasklist.
   If hWnd <> 0 Then
      For i = 0 To m_Count - 1
         If m_tasks(i).hWnd = hWnd Then
            Index = i + 1
            Exit For
         End If
      Next i
   End If
   
   ' Check to see if window still exists.
   If Index > 0 And Index <= m_Count Then
      If IsTask(m_tasks(Index - 1).hWnd) Then
         Valid = True
      Else
         ' Delete that pup!
         m_tasks(Index - 1).hWnd = 0
         m_Count = CompactCache()
         Valid = False
      End If
   End If
End Property

' ************************************************
'  Public Methods
' ************************************************
Public Function Activate(Optional ByVal Index As Long, Optional ByVal hWnd As Long) As Boolean
   Dim i As Long
   
   ' If passed an hWnd, find index within tasklist.
   If hWnd <> 0 Then
      For i = 0 To m_Count - 1
         If m_tasks(i).hWnd = hWnd Then
            Index = i + 1
            Exit For
         End If
      Next i
   End If
   
   ' If we have a valid Index, try to activate.
   If Index > 0 And Index <= m_Count Then
      With m_tasks(Index - 1)
         If Len(.Caption) Then
            If IsIconic(.hWnd) Then
               Call ShowWindow(.hWnd, SW_RESTORE)
            End If
            Call SetForegroundWindow(.hWnd)
            Activate = (GetForegroundWindow = .hWnd)
         End If
      End With
   End If
End Function

Public Sub Refresh()
   Dim i As Long
   Dim hWnd As Long

   ' Flag tasks no longer running
   ' as overwriteable.
   For i = 0 To m_Count - 1
      If IsWindow(m_tasks(i).hWnd) = 0 Then
         m_tasks(i).hWnd = 0
      End If
   Next i
   
   ' Clean up wasted space in cache.
   m_Count = CompactCache()
   
   ' Loop through window list and add
   ' any/all new tasks.
   hWnd = GetTopWindow(0)
   Do While hWnd
      ' Determine if this meets task criteria.
      If IsTask(hWnd) Then Call AddTask(hWnd)
      ' Get next window and move on.
      hWnd = GetNextWindow(hWnd, GW_HWNDNEXT)
   Loop
End Sub

' ************************************************
'  Private Methods
' ************************************************
Private Sub AddTask(ByVal hWnd As Long)
   Dim nIndex As Long
   Dim hWndIcon As Long
   Dim hIcon As Long
   Dim hIconSm As Long
   Dim Caption As String
   Dim lpdwResult As Long
   Dim nLen As Long
   Dim nRet As Long
   Dim i As Long
   Const flags As Long = SMTO_ABORTIFHUNG ' Or SMTO_BLOCK
   
   ' Check if this task is already cached, and
   ' if so store the array Index for future use.
   For i = 0 To m_Count - 1
      If hWnd = m_tasks(i).hWnd Then
         nIndex = i + 1
         Exit For
      End If
   Next i
   
   ' Get length of Caption using SendMessageTimeout
   ' which will return even if app is "hung".
   nRet = SendMessageTimeout(hWnd, WM_GETTEXTLENGTH, 0&, ByVal 0&, flags, m_HungTimeout, lpdwResult)
   
   ' Check if app is hung, and if so bail
   ' if this task is already cached.
   If nRet = 0 Then
      ' In NT, Err.LastDllError will be
      ' set to ERROR_TIMEOUT, but we can't
      ' count on that in Win9x.
      Debug.Print "Timeout on &h" & Hex(hWnd)
      If nIndex Then Exit Sub
   Else
      ' Caption length returned in lpdwResult,
      ' if app isn't hung.
      nLen = lpdwResult + 1
   End If
   
   ' Only tasks with a Caption qualify
   ' for addition to cached task array.
   If nLen > 1 Then
      
      ' Retrieve full Caption.
      Caption = String$(nLen, 0)
      nRet = SendMessageTimeout(hWnd, WM_GETTEXT, nLen, ByVal Caption, flags, m_HungTimeout, lpdwResult)
      If nRet Then
         nLen = lpdwResult
         
         ' If window is owned, then
         ' take icon from owner.
         hWndIcon = GetWindowLong(hWnd, GWL_HWNDPARENT)
         If hWndIcon = 0 Then hWndIcon = hWnd
      
         ' Get the standard hIcon
         If SendMessageTimeout(hWndIcon, WM_GETICON, ICON_BIG, ByVal 0&, flags, m_HungTimeout, lpdwResult) Then
            If lpdwResult Then
               hIcon = lpdwResult
            Else
               hIcon = GetClassLong(hWndIcon, GCL_HICON)
            End If
         End If
         
         ' Get the small hIcon
         If SendMessageTimeout(hWndIcon, WM_GETICON, ICON_SMALL, ByVal 0&, flags, m_HungTimeout, lpdwResult) Then
            If lpdwResult Then
               hIconSm = lpdwResult
            Else
               hIconSm = GetClassLong(hWndIcon, GCL_HICONSM)
            End If
         End If
         
         ' Resize array if no more empty slots.
         If m_Count > UBound(m_tasks) Then
            ReDim Preserve m_tasks(0 To UBound(m_tasks) + ArrayBump) As TaskListItems
         End If
         
         ' Use existing index if task is already
         ' cached, otherwise place at next available
         ' slot in array.
         If nIndex Then
            ' Shift to 0-based index.
            nIndex = nIndex - 1
         Else
            nIndex = m_Count
            ' Bump stored count of tasks.
            m_Count = m_Count + 1
         End If
         
         ' Cache data associated with window.
         m_tasks(nIndex).hWnd = hWnd
         m_tasks(nIndex).hIcon = hIcon
         m_tasks(nIndex).hIconSm = hIconSm
         m_tasks(nIndex).Caption = Left$(Caption, nLen)
      End If
   End If
End Sub

Private Function Classname(ByVal hWnd As Long) As String
   Dim nRet As Long
   Dim Class As String
   Const MaxLen As Long = 256
   
   ' Retrieve classname of passed window.
   Class = String$(MaxLen, 0)
   nRet = GetClassname(hWnd, Class, MaxLen)
   If nRet Then Classname = Left$(Class, nRet)
End Function

Private Function CompactCache() As Long
   Dim i As Long, j As Long
   
   ' Compact cached window array.
   If m_Count Then
      For i = 0 To UBound(m_tasks) - 1
         If m_tasks(i).hWnd = 0 Then
            For j = i To UBound(m_tasks) - 1
               m_tasks(j).hWnd = m_tasks(j + 1).hWnd
               m_tasks(j).hIcon = m_tasks(j + 1).hIcon
               m_tasks(j).hIconSm = m_tasks(j + 1).hIconSm
               m_tasks(j).Caption = m_tasks(j + 1).Caption
            Next j
         End If
      Next i
   End If
   
   ' Get count of cached tasks.
   For i = 0 To UBound(m_tasks)
      If m_tasks(i).hWnd = 0 Then
         CompactCache = i
         Exit For
      End If
   Next i
End Function

Private Function IsTask(ByVal hWnd As Long) As Boolean
   Dim hWndOwner As Long
   
   Select Case m_Mode
      ' Special thanks to Jonathan Wood <http://www.softcircuits.com>
      ' for pointing out that the Alt+Tab order is the same as returned
      ' by a GetWindow loop.
      Case tlAltTab
         ' Must be visible.
         If IsWindowVisible(hWnd) Then
            ' Can't have Parent.
            If GetParent(hWnd) = 0 Then
               ' Can't have Owner *except* in the case where
               ' owner has WS_EX_TOOLWINDOW StyleEx bit set.
               ' This is how VB apps (and others?) operate.
               hWndOwner = GetWindowLong(hWnd, GWL_HWNDPARENT)
               If (hWndOwner = 0) Or IsToolWindow(hWndOwner) Then
                  ' Must not be Toolwindow.
                  If Not IsToolWindow(hWnd) Then
                     ' We have a winner!
                     IsTask = True
                  End If
               End If
            End If
         End If
      
      ' Apply TaskMgr tests to passed window:
      Case tlTaskMgr
         ' Must be visible.
         If IsWindowVisible(hWnd) Then
            ' Can't have Parent.
            If GetParent(hWnd) = 0 Then
               ' And, finally, it shouldn't have an owner.
               If GetWindowLong(hWnd, GWL_HWNDPARENT) = 0 Then
                  ' TaskMgr shows all ToolWindows
                  ' *except* Program Manager.
                  If IsToolWindow(hWnd) Then
                     If Classname(hWnd) <> "Progman" Then
                        ' We have a winner!
                        IsTask = True
                     End If
                  Else
                     ' We have a winner!
                     IsTask = True
                  End If
               End If
            End If
         End If
   End Select
End Function

Private Function IsToolWindow(ByVal hWnd As Long) As Boolean
   ' Check for toolwindow style bit.
   IsToolWindow = ((GetWindowLong(hWnd, GWL_EXSTYLE) And WS_EX_TOOLWINDOW) = WS_EX_TOOLWINDOW)
End Function

